from __future__ import annotations

import libcst as cst
import libcst.matchers as m
from libcst import helpers
from libcst.codemod import CodemodContext
from libcst.codemod.visitors import AddImportsVisitor
from libcst.metadata import ScopeProvider

from django_autotyping.typing import FlattenFunctionDef, ModelType

from .base import InsertAfterImportsVisitor, StubVisitorBasedCodemod
from .constants import OVERLOAD_DECORATOR
from .utils import get_kw_param, get_param

MODEL_T_TYPE_VAR = helpers.parse_template_statement('_ModelT = TypeVar("_ModelT", bound=Model)')
"""A statement assigning `_ModelT = TypeVar("_ModelT", bound=Model)`."""

# Matchers:

RELATED_CLASS_DEF_MATCHER = m.ClassDef(
    name=m.SaveMatchedNode(m.Name("ForeignObject") | m.Name("ForeignKey") | m.Name("OneToOneField"), "field_cls_name"),
)
"""Matches all foreign field class definitions that supports parametrization of the `__set__` and `__get__` types."""

MANY_TO_MANY_CLASS_DEF_MATCHER = m.ClassDef(name=m.Name("ManyToManyField"))
"""Matches the `ManyToManyField` class definition."""

INIT_DEF_MATCHER = m.FunctionDef(name=m.Name("__init__"))
"""Matches the `__init__` method definition."""


class ForwardRelationOverloadCodemod(StubVisitorBasedCodemod):
    """A codemod that will add overloads to the `__init__` methods of related fields.

    **Rule identifier**: `DJAS001`.

    **Related settings**:

    - [`ALLOW_PLAIN_MODEL_REFERENCES`][django_autotyping.app_settings.StubsGenerationSettings.ALLOW_PLAIN_MODEL_REFERENCES]
    - [`ALLOW_NONE_SET_TYPE`][django_autotyping.app_settings.StubsGenerationSettings.ALLOW_NONE_SET_TYPE]

    This will provide auto-completion when using [`ForeignKey`][django.db.models.ForeignKey],
    [`OneToOneField`][django.db.models.OneToOneField] and [`ManyToManyField`][django.db.models.ManyToManyField]
    with string references to a model, and accurate type checking when accessing the field attribute
    from a model instance.

    ```python
    class MyModel(models.Model):
        field = models.ForeignKey(
            "myapp.Other",
            on_delete=models.CASCADE,
        )
        nullable = models.OneToOneField(
            "myapp.Other",
            on_delete=models.CASCADE,
            null=True,
        )
    reveal_type(MyModel().field)  # Revealed type is "Other"
    reveal_type(MyModel().nullable)  # Revealed type is "Other | None"
    ```

    !!! info "Usage with VSCode"
        Auto-completion might not always work as expected, see this
        [issue](https://github.com/microsoft/pylance-release/issues/4428).


    ??? abstract "Implementation"
        The following is a snippet of the produced overloads:

        ```python
        class ForeignKey(ForeignObject[_ST, _GT]):
            # For each model, will add two overloads:
            # - 1st: `null: Literal[True]`, which will parametrize `ForeignKey` types as `Optional`.
            # - 2nd: `null: Literal[False] = ...` (the default).
            # `to` is annotated as a `Literal`, with two values: {app_label}.{model_name} and {model_name}.
            @overload
            def __init__(
                self: ForeignKey[MyModel | Combinable | None, MyModel | None],
                to: Literal["MyModel", "myapp.MyModel"],
                ...
            ) -> None: ...
        ```
    """  # noqa: E501

    METADATA_DEPENDENCIES = {ScopeProvider}
    STUB_FILES = {"db/models/fields/related.pyi"}

    def __init__(self, context: CodemodContext) -> None:
        super().__init__(context)
        self.add_model_imports()
        InsertAfterImportsVisitor.insert_after_imports(context, [MODEL_T_TYPE_VAR])

        # Even though these are most likely included, we import them for safety:
        self.add_typing_imports(["Literal", "TypeVar", "overload"])

        AddImportsVisitor.add_needed_import(
            self.context,
            module="django.db.models.expressions",
            obj="Combinable",
        )

    @m.call_if_inside(MANY_TO_MANY_CLASS_DEF_MATCHER)
    @m.leave(INIT_DEF_MATCHER)
    def mutate_ManyToManyField_FunctionDef(
        self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
    ) -> FlattenFunctionDef:
        """Add the necessary overloads for `ManyToManyField`'s `__init__` method.

        Due to combinatorial explosion, we can't add overloads that would handle `through` models alongside with `to`.
        """
        overload_init = updated_node.with_changes(decorators=[OVERLOAD_DECORATOR])
        overloads: list[cst.FunctionDef] = []

        for model in self.django_context.models:
            model_name = self.django_context.get_model_name(model)
            allow_plain_model_name = (
                self.stubs_settings.ALLOW_PLAIN_MODEL_REFERENCES and not self.django_context.is_duplicate(model)
            )

            # sets `self: ManyToManyField[model_name, _Through]`
            self_param = get_param(overload_init, "self")
            overload = overload_init.with_deep_changes(
                old_node=self_param,
                annotation=cst.Annotation(
                    annotation=helpers.parse_template_expression(f"ManyToManyField[{model_name}, _Through]")
                ),
            )

            # sets `to: Literal["model_name", "app_label.model_name"]`
            # (or just "app_label.model_name" if duplicate model names)
            to_param = get_param(overload, "to")
            overload = overload.with_deep_changes(
                old_node=to_param,
                annotation=_build_to_annotation(model, allow_plain_model_name),
            )

            overloads.append(overload)

        # Now, handle the last overload, matching against a real model type:

        # sets `to: type[_To]`, essentially removing the `| str` part. This way,
        # we don't need to explicitly annotate `self`, type checkers will infer this natively.
        to_param = get_param(overload_init, "to")
        model_overload = overload_init.with_deep_changes(
            old_node=to_param,
            annotation=cst.Annotation(
                annotation=cst.Subscript(
                    value=cst.Name("type"), slice=[cst.SubscriptElement(cst.Index(cst.Name("_To")))]
                )
            ),
        )
        overloads.append(model_overload)

        return cst.FlattenSentinel(overloads)

    @m.call_if_inside(RELATED_CLASS_DEF_MATCHER)
    @m.leave(INIT_DEF_MATCHER)
    def mutate_relatedFields_FunctionDef(
        self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
    ) -> FlattenFunctionDef:
        """Add the necessary overloads to foreign fields that supports
        that supports parametrization of the `__set__` and `__get__` types.
        """
        field_cls_name = self.get_metadata(ScopeProvider, original_node).name

        overload_init = updated_node.with_changes(decorators=[OVERLOAD_DECORATOR])
        overloads: list[cst.FunctionDef] = []

        # For each model, create two overloads, depending on the `null` value:
        for model in self.django_context.models:
            model_name = self.django_context.get_model_name(model)
            allow_plain_model_name = (
                self.stubs_settings.ALLOW_PLAIN_MODEL_REFERENCES and not self.django_context.is_duplicate(model)
            )

            for nullable in (True, False):  # Order matters!
                # sets `self: FieldName[<set_type>, <get_type>]`
                self_param = get_param(overload_init, "self")
                overload = overload_init.with_deep_changes(
                    old_node=self_param,
                    annotation=_build_self_annotation(
                        field_cls_name, model_name, nullable, self.stubs_settings.ALLOW_NONE_SET_TYPE
                    ),
                )

                # sets `to: Literal["model_name", "app_label.model_name"]`
                to_param = get_param(overload, "to")
                overload = overload.with_deep_changes(
                    old_node=to_param,
                    annotation=_build_to_annotation(model, allow_plain_model_name),
                )

                # sets `null: Literal[True/False]` (with the default removed accordingly)
                null_param = get_kw_param(overload, "null")
                overload = overload.with_deep_changes(
                    old_node=null_param,
                    annotation=cst.Annotation(
                        annotation=cst.Subscript(
                            value=cst.Name("Literal"), slice=[cst.SubscriptElement(cst.Index(cst.Name(str(nullable))))]
                        )
                    ),
                    default=None if nullable else null_param.default,  # Remove default to have a correct overload match
                    equal=cst.MaybeSentinel.DEFAULT if nullable else null_param.equal,
                )

                overloads.append(overload)

        # Now, handle the last overload, matching against a real model type:

        # sets `to: type[_ModelT]`, the type variable will be used to annotate `self` as well
        to_param = get_param(overload_init, "to")
        model_overload = overload_init.with_deep_changes(
            old_node=to_param,
            annotation=cst.Annotation(
                annotation=cst.Subscript(
                    value=cst.Name("type"), slice=[cst.SubscriptElement(cst.Index(cst.Name("_ModelT")))]
                )
            ),
        )

        for nullable in (True, False):  # Order matters!
            # sets `self: FieldName[<set_type>, <get_type>]`
            self_param = get_param(model_overload, "self")
            model_overload_ = model_overload.with_deep_changes(
                old_node=self_param,
                annotation=_build_self_annotation(
                    field_cls_name, "_ModelT", nullable, self.stubs_settings.ALLOW_NONE_SET_TYPE
                ),
            )

            # sets `null: Literal[True/False]` (with the default removed accordingly)
            null_param = get_kw_param(model_overload_, "null")
            model_overload_ = model_overload_.with_deep_changes(
                old_node=null_param,
                annotation=cst.Annotation(
                    annotation=cst.Subscript(
                        value=cst.Name("Literal"), slice=[cst.SubscriptElement(cst.Index(cst.Name(str(nullable))))]
                    )
                ),
                default=None if nullable else null_param.default,  # Remove default to have a correct overload match
                equal=cst.MaybeSentinel.DEFAULT if nullable else null_param.equal,
            )

            overloads.append(model_overload_)

        # Temp workaround to have autocompletion working, this overload shouldn't be used as a match by type checkers
        # to_param = get_param(overload_init, "to")
        # literal_completion_overload = overload_init.with_deep_changes(
        #     old_node=to_param,
        #     annotation=cst.Annotation(
        #         annotation=cst.Subscript(
        #             value=cst.Name("Literal"),
        #             slice=[
        #                 cst.SubscriptElement(cst.Index(cst.SimpleString(string)))
        #                 for model in self.django_models
        #                 for string in (f'"{model.__name__}"', f'"{model._meta.app_label}.{model.__name__}"')
        #             ],
        #         )
        #     ),
        # )

        # overloads.append(literal_completion_overload)

        return cst.FlattenSentinel(overloads)


def _build_self_annotation(
    field_cls_name: str, model_name: str, nullable: bool, allow_none_set_type: bool
) -> cst.Annotation:
    """Builds the `self` annotation of foreign fields.

    With `field_cls_name="ForeignKey"`, `model_name="MyModel"` and `nullable=False`, the following is produced:

    >>> ForeignKey[MyModel | None, MyModel]

    (Even if not nullable, the `__set__` type can still be `None`. Having a foreign instance is only enforced on save).
    """
    set_type = f"{model_name} | Combinable | None" if allow_none_set_type or nullable else f"{model_name} | Combinable"
    get_type = f"{model_name} | None" if nullable else model_name
    return cst.Annotation(annotation=helpers.parse_template_expression(f"{field_cls_name}[{set_type}, {get_type}]"))


def _build_to_annotation(model: ModelType, allow_plain_model_name: bool) -> cst.Annotation:
    """Builds the `to` annotation of foreign fields.

    This will result in a `Literal` with two string values, the model name and the dotted app label and model name.
    If `allow_plain_model_name` is set to `False`, only the second literal value will be set.
    """
    slice = [cst.SubscriptElement(cst.Index(cst.SimpleString(f'"{model._meta.app_label}.{model.__name__}"')))]
    if allow_plain_model_name:
        slice.insert(0, cst.SubscriptElement(cst.Index(cst.SimpleString(f'"{model.__name__}"'))))

    return cst.Annotation(
        annotation=cst.Subscript(
            value=cst.Name("Literal"),
            slice=slice,
        )
    )



from __future__ import annotations

import difflib
from pathlib import Path
from typing import Any, Iterable

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand, CommandParser
from django.utils.termcolors import colorize

from django_autotyping._compat import Unpack
from django_autotyping.app_settings import AutotypingSettings
from django_autotyping.codemodding.codemods import RulesT, gather_codemods, rules
from django_autotyping.codemodding.django_context import DjangoCodemodContext
from django_autotyping.codemodding.main import run_codemods

from ._utils import BaseOptions, dir_path

at_settings = AutotypingSettings(settings)


class CommandOptions(BaseOptions):
    project_dir: Path
    diff: bool
    ignore: list[RulesT] | None


class Command(BaseCommand):
    help = "Add type annotations to your Django code."

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "-p",
            "--project-dir",
            type=dir_path,
            help="The directory of the project, where code modifications should be applied.",
            required=at_settings.CODE_GENERATION.PROJECT_DIR is None,
            default=at_settings.CODE_GENERATION.PROJECT_DIR,
        )
        parser.add_argument(
            "--diff",
            action="store_true",
            help="Show changes to be applied instead of modifying existing files.",
            default=at_settings.CODE_GENERATION.DIFF,
        )
        parser.add_argument(
            "--ignore",
            choices=[rule[0] for rule in rules],
            nargs="*",
            help="Rules to be ignored.",
            default=at_settings.IGNORE,
        )

    def _colored_diff(self, lines: Iterable[str]) -> None:
        for line in lines:
            line_s = line.rstrip("\n")
            if line_s.startswith("+"):
                self.stdout.write(colorize(line_s, fg="green"))
            elif line_s.startswith("-"):
                self.stdout.write(colorize(line_s, fg="red"))
            elif line_s.startswith("^"):
                self.stdout.write(colorize(line_s, fg="blue"))
            else:
                self.stdout.write(line_s)

    def handle(self, *args: Any, **options: Unpack[CommandOptions]) -> None:
        django_context = DjangoCodemodContext(apps, settings, options["project_dir"])
        codemods = gather_codemods(options["ignore"])

        # TODO codemods should specify which type of file they apply to.
        model_filenames = set(model_info.filename for model_info in django_context.model_infos)

        for filename in model_filenames:
            intput_source = Path(filename).read_text("utf-8")
            output_source = run_codemods(codemods, django_context, at_settings.CODE_GENERATION, filename)
            if intput_source != output_source:
                if options["diff"]:
                    lines = difflib.unified_diff(
                        intput_source.splitlines(keepends=True),
                        output_source.splitlines(keepends=True),
                        fromfile=filename,
                        tofile=filename,
                    )
                    self._colored_diff(lines)
                else:
                    Path(filename).write_text(output_source, encoding="utf-8")



from __future__ import annotations

from pathlib import Path
from typing import Any

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand, CommandParser

from django_autotyping._compat import Unpack
from django_autotyping.app_settings import AutotypingSettings
from django_autotyping.stubbing import create_local_django_stubs, run_codemods
from django_autotyping.stubbing.codemods import RulesT, gather_codemods, rules
from django_autotyping.stubbing.django_context import DjangoStubbingContext

from ._utils import BaseOptions, dir_path

at_settings = AutotypingSettings.from_django_settings(settings)
stubs_settings = at_settings.STUBS_GENERATION


class CommandOptions(BaseOptions):
    local_stubs_dir: Path
    ignore: list[RulesT] | None


class Command(BaseCommand):
    help = "Generate customized type stubs for your Django application."

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "-s",
            "--local-stubs-dir",
            type=dir_path,
            help="The directory of the local type stubs.",
            required=at_settings.STUBS_GENERATION.LOCAL_STUBS_DIR is None,
            default=at_settings.STUBS_GENERATION.LOCAL_STUBS_DIR,
        )
        parser.add_argument(
            "--ignore",
            choices=[rule[0] for rule in rules],
            nargs="*",
            help="Rules to be ignored.",
            default=at_settings.IGNORE,
        )

    def handle(self, *args: Any, **options: Unpack[CommandOptions]) -> None:
        create_local_django_stubs(options["local_stubs_dir"], stubs_settings.SOURCE_STUBS_DIR)
        codemods = gather_codemods(options["ignore"])

        django_context = DjangoStubbingContext(apps, settings)
        run_codemods(codemods, django_context, stubs_settings)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

import libcst as cst
import libcst.matchers as m
from libcst import helpers
from libcst.codemod import CodemodContext

from django_autotyping.typing import FlattenFunctionDef

from .base import InsertAfterImportsVisitor, StubVisitorBasedCodemod
from .constants import OVERLOAD_DECORATOR
from .utils import TypedDictAttribute, build_typed_dict, get_param

if TYPE_CHECKING:
    from ..django_context import PathInfo

# `SupportsStr` is a Protocol that supports `__str__`.
# This should be equivalent to `object`, but is used to be
# more explicit.
SUPPORTS_STR_DEF = helpers.parse_template_statement(
    """
class SupportsStr(Protocol):
    def __str__(self) -> str:
        ...
"""
)

EMPTY_DICT_DEF = helpers.parse_template_statement(
    """
class EmptyDict(TypedDict):
    pass
"""
)

LITERAL_NONE = helpers.parse_template_expression("Literal[None]")

# Matchers:

REVERSE_DEF_MATCHER = m.FunctionDef(name=m.Name("reverse"))
"""Matches the `reverse` function definition."""


class ReverseOverloadCodemod(StubVisitorBasedCodemod):
    """A codemod that will add overloads to the [`reverse`][django.urls.reverse] function.

    **Rule identifier**: `DJAS011`.

    **Related settings**:

    - [`ALLOW_REVERSE_ARGS`][django_autotyping.app_settings.StubsGenerationSettings.ALLOW_REVERSE_ARGS].

    ```python
    reverse("my-view-name", kwargs={...})  # `kwargs` is typed with a `TypedDict`, providing auto-completion.
    ```
    """

    STUB_FILES = {"urls/base.pyi"}

    def __init__(self, context: CodemodContext) -> None:
        super().__init__(context)

        InsertAfterImportsVisitor.insert_after_imports(context, [EMPTY_DICT_DEF, SUPPORTS_STR_DEF])
        self.add_typing_imports(["Literal", "TypedDict", "NotRequired", "Protocol", "overload"])

    @m.leave(REVERSE_DEF_MATCHER)
    def mutate_FunctionDef(self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef) -> FlattenFunctionDef:
        overload = updated_node.with_changes(decorators=[OVERLOAD_DECORATOR])

        overloads: list[cst.FunctionDef] = []
        seen_typeddict_names: list[str] = []
        reversed_dict: defaultdict[PathInfo, list[str]] = defaultdict(list)

        # First, build a reverse dictionary: a mapping between PathInfos instances (shared between views)
        # and a list of viewnames

        for viewname, path_info in self.django_context.viewnames_lookups.items():
            reversed_dict[path_info].append(viewname)

        for path_info, viewnames in reversed_dict.items():
            # We do not support `current_app` for now, it would generate too many overloads
            overload_ = overload.with_deep_changes(
                old_node=get_param(overload, "current_app"),
                annotation=cst.Annotation(LITERAL_NONE),
            )

            viewnames_literals = ", ".join(f'"{viewname}"' for viewname in viewnames)
            overload_ = overload_.with_deep_changes(
                old_node=get_param(overload_, "viewname"),
                annotation=cst.Annotation(helpers.parse_template_expression(f"Literal[{viewnames_literals}]")),
            )

            if path_info.is_empty:
                # Calling `reverse` with `args` or `kwargs` will fail at runtime if the view has no arguments.
                # We create a special overload handling this case:
                overload_ = overload_.with_deep_changes(
                    old_node=get_param(overload_, "args"),
                    annotation=cst.Annotation(helpers.parse_template_expression("tuple[()] | None")),
                )
                overload_ = overload_.with_deep_changes(
                    old_node=get_param(overload_, "kwargs"),
                    annotation=cst.Annotation(helpers.parse_template_expression("EmptyDict | None")),
                )
                overloads.insert(0, overload_)
                continue

            use_args_options = (True, False) if self.stubs_settings.ALLOW_REVERSE_ARGS else (False,)

            for use_args in use_args_options:
                args_param = get_param(overload_, "args")
                if use_args:
                    annotation = helpers.parse_template_expression(path_info.get_args_annotation())
                else:
                    annotation = LITERAL_NONE

                args_param = args_param.with_changes(
                    annotation=cst.Annotation(annotation),
                    default=None if use_args else args_param.default,
                    equal=cst.MaybeSentinel.DEFAULT if use_args else args_param.equal,
                )

                kwargs_param = get_param(overload_, "kwargs")
                if use_args:
                    annotation = LITERAL_NONE
                else:
                    annotation = helpers.parse_template_expression(path_info.get_kwargs_annotation())

                    # Add the TypedDict definition if not already done:
                    for path_args in path_info.arguments_set:
                        typeddict_name = path_args.typeddict_name
                        if typeddict_name in seen_typeddict_names:
                            continue

                        seen_typeddict_names.append(typeddict_name)
                        typed_dict = build_typed_dict(
                            typeddict_name,
                            attributes=[
                                TypedDictAttribute(
                                    name=arg_name,
                                    annotation="SupportsStr",
                                    not_required=True if not required else None,
                                    # TODO, any docstring?
                                )
                                for arg_name, required in path_args.arguments
                            ],
                            leading_line=True,
                        )

                        InsertAfterImportsVisitor.insert_after_imports(self.context, [typed_dict])

                kwargs_param = kwargs_param.with_changes(
                    annotation=cst.Annotation(annotation),
                    default=None if not use_args else kwargs_param.default,
                    equal=cst.MaybeSentinel.DEFAULT if not use_args else kwargs_param.equal,
                )

                # Finally, add a `ParamStar` after `urlconf`, to have valid signatures.
                # Also move the necessary arguments as kwonly_params:

                overload_param_star = overload_.with_deep_changes(
                    old_node=overload_.params,
                    star_arg=cst.ParamStar(),
                    params=[p for p in overload_.params.params if p.name.value in ("viewname", "urlconf")],
                    kwonly_params=[args_param, kwargs_param, get_param(overload_, "current_app")],
                )

                overloads.append(overload_param_star)

        # Remove the `str` annotation from `viewname` in the fallback overloads, so that
        # only literals will match:
        overload = overload.with_deep_changes(
            old_node=get_param(overload, "viewname"),
            annotation=cst.Annotation(helpers.parse_template_expression("Callable[..., Any] | None")),
        )

        return cst.FlattenSentinel(overloads + [overload])



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path

from django.conf import LazySettings

from ._compat import Self
from .typing import AutotypingSettingsDict, RulesT


@dataclass
class CodeGenerationSettings:
    """Configuration for adding type annotations to Django user code."""

    PROJECT_DIR: Path | None = None
    """The directory of the project, where code modifications should be applied."""

    DIFF: bool = False
    """Show changes to be applied instead of modifying existing files."""

    TYPE_CHECKING_BLOCK: bool = True
    """Whether newly added imports should be in an `if TYPE_CHECKING` block (avoids circular imports)."""

    ASSUME_CLASS_GETITEM: bool = False
    """Whether generic classes in stubs files but not at runtime should be assumed to have a
    `__class_getitem__` method. This can be achieved by using `django-stubs-ext` or manually.

    Affected rules: `DJA001`.
    """


@dataclass
class StubsGenerationSettings:
    """Configuration for dynamic stubs generation."""

    LOCAL_STUBS_DIR: Path | None = None
    """The directory of the local type stubs. If not set, this setting must be set as a CLI argument."""

    SOURCE_STUBS_DIR: Path | None = None
    """The directory of the source `django-stubs` to be used. Will default
    to the first entry in site packages.
    """

    ALLOW_PLAIN_MODEL_REFERENCES: bool = True
    """Whether string references in the form of `{model_name}` should be generated in overloads.

    If set to `True`, both `{model_name}` and `{model_name}.{app_label}` are allowed
    (unless the model name has a duplicate in a different app).

    Affected rules: `DJAS001`.
    """

    ALLOW_NONE_SET_TYPE: bool = False
    """Whether to allow having the `__set__` type variable set to `None`, even if the field is not nullable.

    While Django allows setting most model instance fields to any value (before saving),
    it is generally a bad practice to do so. However, it might be beneficial to allow `None`
    to be set temporarly.

    This also works for foreign fields, where unlike standard fields, the Django descriptor used
    only allows model instances and `None` to be set.

    Affected rules: `DJAS001`.
    """

    MODEL_FIELDS_OPTIONAL: bool = True
    """Whether all model fields should be considered optional when creating model instances.

    This affects the following signatures:

    - [`Manager.create/acreate`][django.db.models.Manager]
    - `__init__` methods of models

    A lot can happen behind the scenes when instantiating models. Even if a field doesn't have
    a default value provided, the database could have triggers implemented that would provide one.
    This is why, by default, this configuration attribute defaults to `True`. If set to `False`,
    `django-autotyping` will try its best to determine required fields, namely by checking if:

    - the field can be [`null`][django.db.models.Field.null]
    - the field has a default or a database default value set
    - the field is a subclass of [`DateField`][django.db.models.DateField] and has
      [`auto_now`][django.db.models.DateField.auto_now] or [`auto_now_add`][django.db.models.DateField.auto_now_add]
      set to `True`.

    Affected rules: `DJAS002`.
    """

    ALLOW_REVERSE_ARGS: bool = False
    """Whether type checking should be added to the `args` argument of [`reverse`][django.urls.reverse].

    By default, this is set to `False` to avoid having too many overloads being generated.
    Moreover, only tuples can be type checked, and most people are using lists for this argument.
    Instead, it is recommended to use the `kwargs` argument.

    Affected rules: `DJAS011`.
    """


@dataclass
class AutotypingSettings:
    """A class holding the django-autotyping configuration."""

    IGNORE: list[RulesT] = field(default_factory=list)
    """A list of ignored rules."""

    STUBS_GENERATION: StubsGenerationSettings = field(default_factory=StubsGenerationSettings)
    """Stub related settings."""

    CODE_GENERATION: CodeGenerationSettings = field(default_factory=CodeGenerationSettings)
    """Code generation related settings."""

    @classmethod
    def from_django_settings(cls, settings: LazySettings) -> Self:
        # Not pretty, but we are limited by dataclasses
        autotyping_settings: AutotypingSettingsDict = deepcopy(getattr(settings, "AUTOTYPING", {}))
        stubs_generation_dct = autotyping_settings.pop("STUBS_GENERATION", {})
        code_generation_dct = autotyping_settings.pop("CODE_GENERATION", {})
        return cls(
            **autotyping_settings,
            STUBS_GENERATION=StubsGenerationSettings(**stubs_generation_dct),
            CODE_GENERATION=CodeGenerationSettings(**code_generation_dct),
        )



from __future__ import annotations

from pathlib import Path

import libcst as cst
from libcst.codemod import CodemodContext, VisitorBasedCodemodCommand

from django_autotyping.app_settings import CodeGenerationSettings

from .django_context import DjangoCodemodContext
from .vendoring.monkeytype import MoveImportsToTypeCheckingBlockVisitor, get_newly_imported_items


def run_codemods(
    codemods: list[type[VisitorBasedCodemodCommand]],
    django_context: DjangoCodemodContext,
    code_generation_settings: CodeGenerationSettings,
    filename: str,
) -> str:
    context = CodemodContext(
        filename=filename,
        scratch={
            "django_context": django_context,
            "code_generation_settings": code_generation_settings,
        },
    )

    input_code = Path(filename).read_text(encoding="utf-8")
    input_module = cst.parse_module(input_code)
    output_module = cst.parse_module(input_code)
    for codemod in codemods:
        transformer = codemod(context=context)

        output_module = transformer.transform_module(output_module)

    if code_generation_settings.TYPE_CHECKING_BLOCK:
        newly_imported_items = get_newly_imported_items(output_module, input_module)
        if newly_imported_items:
            context = CodemodContext()
            MoveImportsToTypeCheckingBlockVisitor.store_imports_in_context(
                context,
                newly_imported_items,
            )
            type_checking_block_transformer = MoveImportsToTypeCheckingBlockVisitor(context)
            output_module = type_checking_block_transformer.transform_module(output_module)

    return output_module.code



from __future__ import annotations

from argparse import ArgumentTypeError
from pathlib import Path
from typing import Literal, TypedDict


def dir_path(path_str: str) -> Path:
    if path_str and not (path := Path(path_str)).is_dir():
        raise ArgumentTypeError(f"{path_str} must be an existing directory.")
    return path


class BaseOptions(TypedDict):
    verbosity: Literal[0, 1, 2, 3]
    """Verbosity level; 0=minimal output, 1=normal output, 2=verbose output, 3=very verbose output"""

    settings: str | None
    """The Python path to a settings module, e.g. "myproject.settings.main". If this isn't provided,
    the DJANGO_SETTINGS_MODULE environment variable will be used.
    """

    pythonpath: str | None
    """A directory to add to the Python path, e.g. "/home/djangoprojects/myproject"."""

    traceback: bool
    """Raise on CommandError exceptions."""

    no_color: bool
    """Don't colorize the command output."""

    skip_checks: bool
    """Skip system checks."""



from __future__ import annotations

import re
from dataclasses import dataclass

import libcst as cst
from libcst import helpers
from libcst import matchers as m


def get_method_node(class_node: cst.ClassDef, method_name: str) -> cst.FunctionDef:
    method_def = m.FunctionDef(name=m.Name(method_name))
    return helpers.ensure_type(
        next(node for node in class_node.body.body if m.matches(node, method_def)), cst.FunctionDef
    )


def get_param(node: cst.FunctionDef, param_name: str) -> cst.Param:
    """Get the `Param` node matching `param_name`."""
    try:
        return next(param for param in node.params.params if param.name.value == param_name)
    except StopIteration:
        raise RuntimeError(
            f"The `FunctionDef` node with name {node.name.value!r} does not have any parameter named {param_name!r}"
        )


def get_kw_param(node: cst.FunctionDef, param_name: str) -> cst.Param:
    """Get the keyword only `Param` node matching `param_name`."""
    return next(param for param in node.params.kwonly_params if param.name.value == param_name)


def to_pascal(string) -> str:
    return re.sub("([0-9A-Za-z])_(?=[0-9A-Z])", lambda m: m.group(1), string.title())


def _indent(string: str) -> str:
    return string.replace("\n", "\n    ")


@dataclass
class TypedDictAttribute:
    name: str
    """The attribute name."""

    annotation: str
    """The annotation of the field."""

    docstring: str | None = None
    """The docstring of the field."""

    required: bool = False
    """Whether the field should be marked as `Required`."""

    not_required: bool = False
    """Whether the field should be marked as `NotRequired`."""

    @property
    def marked_annotation(self) -> str:
        """The annotation additionally marked as required or not required."""
        if self.required:
            return f"Required[{self.annotation}]"
        if self.not_required:
            return f"NotRequired[{self.annotation}]"
        return self.annotation

    def __post_init__(self):
        if self.required and self.not_required:
            raise ValueError("`required` and `not_required` can't be set together.")


def build_typed_dict(
    name: str, attributes: list[TypedDictAttribute], total: bool = True, leading_line: bool = False
) -> cst.SimpleStatementLine | cst.ClassDef:
    """Build a `TypedDict` class definition.

    If one of the attribute's name is not a valid Python identifier, the alternative functional syntax
    will be used (a `SimpleStatementLine` will be created instead of a `ClassDef`).

    Args:
        name: The name of the resulting class.
        attributes: A list of `TypedDictAttribute` instances, representing attributes of the dict.
        total: Whether `total=True` should be used.
        leadind_line: Whether an empty leading line should be added before the class definition.

    """
    functional = any(not attr.name.isidentifier() for attr in attributes)
    leading_lines = [cst.EmptyLine(indent=False)] if leading_line else []
    if not functional:
        body: list[cst.SimpleStatementLine] = []

        for attr in attributes:
            ann_statement = helpers.parse_template_statement(f"{attr.name}: {attr.marked_annotation}")
            if attributes.index(attr) != 0:
                ann_statement = ann_statement.with_changes(leading_lines=[cst.EmptyLine(indent=False)])
            body.append(ann_statement)

            if attr.docstring:
                docstring = f'"""{_indent(attr.docstring)}"""'
                body.append(cst.SimpleStatementLine(body=[cst.Expr(cst.SimpleString(docstring))]))

        return cst.ClassDef(
            name=cst.Name(name),
            bases=[cst.Arg(cst.Name("TypedDict"))],
            keywords=[
                cst.Arg(
                    keyword=cst.Name("total"),
                    equal=cst.AssignEqual(cst.SimpleWhitespace(""), cst.SimpleWhitespace("")),
                    value=cst.Name("False"),
                )
            ]
            if not total
            else [],
            body=cst.IndentedBlock(body),
            leading_lines=leading_lines,
        )

    # If some attributes aren't Python identifiers, we use the functional form:
    # name = TypedDict("name", {"x": int, "y": int})
    return cst.SimpleStatementLine(
        body=[
            cst.Assign(
                targets=[cst.AssignTarget(cst.Name(name))],
                value=cst.Call(
                    func=cst.Name("TypedDict"),
                    args=[
                        cst.Arg(cst.SimpleString(f'"{name}"')),
                        cst.Arg(
                            cst.Dict(
                                elements=[
                                    cst.DictElement(
                                        key=cst.SimpleString(f'"{attr.name}"'), value=cst.Name(attr.marked_annotation)
                                    )
                                    for attr in attributes
                                ]
                            )
                        ),
                    ],
                ),
            )
        ],
        leading_lines=leading_lines,
    )



from __future__ import annotations

from typing import Container, Literal

from django_autotyping._compat import TypeAlias

from .base import StubVisitorBasedCodemod
from .create_overload_codemod import CreateOverloadCodemod
from .forward_relation_overload_codemod import ForwardRelationOverloadCodemod
from .get_model_overload_codemod import GetModelOverloadCodemod
from .query_lookups_overload_codemod import QueryLookupsOverloadCodemod
from .reverse_overload_codemod import ReverseOverloadCodemod

__all__ = (
    "StubVisitorBasedCodemod",
    "CreateOverloadCodemod",
    "ForwardRelationOverloadCodemod",
    "GetModelOverloadCodemod",
    "QueryLookupsOverloadCodemod",
    "ReverseOverloadCodemod",
)

RulesT: TypeAlias = Literal["DJAS001", "DJAS002", "DJAS010", "DJAS011"]

rules: list[tuple[RulesT, type[StubVisitorBasedCodemod]]] = [
    ("DJAS001", ForwardRelationOverloadCodemod),
    ("DJAS002", CreateOverloadCodemod),
    # ("DJAS003", QueryLookupsOverloadCodemod),
    ("DJAS010", GetModelOverloadCodemod),
    ("DJAS011", ReverseOverloadCodemod),
]


def gather_codemods(
    ignore: Container[RulesT] = [], include: Container[RulesT] = []
) -> list[type[StubVisitorBasedCodemod]]:
    if include:
        return [rule[1] for rule in rules if rule[0] in include]
    return [rule[1] for rule in rules if rule[0] not in ignore]



from __future__ import annotations

import hashlib
from collections import defaultdict
from dataclasses import dataclass, field, replace

from django.apps.registry import Apps
from django.conf import LazySettings
from django.db.models import NOT_PROVIDED, DateField, Field
from django.urls import URLPattern, URLResolver, get_resolver
from libcst.codemod.visitors import ImportItem

from django_autotyping.typing import ModelType

from .codemods.utils import to_pascal


@dataclass(eq=True, frozen=True)
class PathArguments:
    """Describes the available arguments for a specific Django view."""

    arguments: frozenset[tuple[str, bool]] = field(default_factory=frozenset)

    def __len__(self) -> int:
        return len(self.arguments)

    def __bool__(self) -> bool:
        return bool(len(self))

    @property
    def sha1(self) -> str:
        stringified = "".join(f"{k}={v}" for k, v in sorted(self.arguments, key=lambda arg: arg[0]))
        return hashlib.sha1(stringified.encode("utf-8")).hexdigest()

    @property
    def typeddict_name(self) -> str:
        return f"_{self.sha1[:6].upper()}Kwargs"

    def is_mergeable(self, arguments: dict[str, bool]) -> bool:
        """Return whether the keys of the provided arguments are the same as the current instance."""
        return set(arg[0] for arg in self.arguments) == set(arguments)

    def with_new_arguments(self, arguments: dict[str, bool]) -> PathArguments:
        new_arguments = frozenset((k, False if not arguments[k] else is_required) for k, is_required in self.arguments)
        return replace(self, arguments=new_arguments)


@dataclass(eq=True, frozen=True)
class PathInfo:
    """Describes the set of available arguments for a Django view.

    At its core, this class holds a set of the possible combinations of arguments for a view.
    If multiple views are registered with the same name, a call to `reverse` can match any of
    these views depending on the provided arguments and the order they appear in the URL patterns list.
    """

    arguments_set: frozenset[PathArguments] = field(default_factory=frozenset)

    @property
    def is_empty(self) -> bool:
        return all(not args for args in self.arguments_set)

    def with_new_arguments(self, arguments: dict[str, bool]) -> PathInfo:
        unfrozen_set = set(self.arguments_set)

        for args in self.arguments_set:
            if args.is_mergeable(arguments):
                new_args = args.with_new_arguments(arguments)
                unfrozen_set.remove(args)
                unfrozen_set.add(new_args)
                break
        else:
            # Provided arguments aren't mergeable, add a new entry
            unfrozen_set.add(PathArguments(frozenset(arguments.items())))
        return replace(self, arguments_set=frozenset(unfrozen_set))

    def get_kwargs_annotation(self) -> str:
        """Return the type annotation for the `kwargs` argument of `reverse`."""
        tds_str = [args.typeddict_name for args in self.arguments_set if args]
        if any(not args for args in self.arguments_set):
            tds_str.extend(["EmptyDict", "None"])

        return " | ".join(tds_str)

    def get_args_annotation(self, list_fallback: bool = True) -> str:
        """Return the type annotation for the `args` argument of `reverse`.

        If multiple URL patterns are available for a specific URL name (i.e. `arguments_set` contains
        multiple entries), the generated annotation will be the union of the possible tuple shapes.

        Args:
            list_fallback: Whether to include a `list[Any]` fallback type.
        """
        args_lengths = sorted([len(args) for args in self.arguments_set], reverse=True)
        tuples_str = [
            f"tuple[{', '.join('SupportsStr' for _ in range(length))}]" for length in args_lengths if length > 0
        ]
        if 0 in args_lengths:
            tuples_str.append("tuple[()]")
        if list_fallback:
            tuples_str.append("list[Any]")
        if 0 in args_lengths:
            tuples_str.append("None")

        return " | ".join(tuples_str)


class DjangoStubbingContext:
    def __init__(self, apps: Apps, settings: LazySettings) -> None:
        self.apps = apps
        self.settings = settings

    @staticmethod
    def _get_model_alias(model: ModelType) -> str:
        """Return an alias of the model, by converting the app label to PascalCase and joining
        the app label to the model name.
        """
        app_label = to_pascal(model._meta.app_label)
        return f"{app_label}{model.__name__}"

    @property
    def models(self) -> list[ModelType]:
        """All the defined models."""
        return self.apps.get_models()

    @property
    def model_imports(self) -> list[ImportItem]:
        """A list of `ImportItem` instances.

        Can be used to easily import all models in a stub file.
        """

        return [
            ImportItem(
                module_name=model._meta.app_config.models_module.__name__,
                obj_name=model.__name__,
                alias=self._get_model_alias(model) if self.is_duplicate(model) else None,
            )
            for model in self.models
        ]

    @property
    def viewnames_lookups(self) -> defaultdict[str, PathInfo]:
        """A mapping between viewnames to be used with `reverse` and the available lookup arguments."""
        return _get_paths_infos(get_resolver())

    def is_duplicate(self, model: ModelType) -> bool:
        """Whether the model has a duplicate name with another model in a different app."""
        return len([m for m in self.models if m.__name__ == model.__name__]) >= 2  # noqa: PLR2004

    def get_model_name(self, model: ModelType) -> str:
        """Return the name of the model in the context of a stub file.

        If the model has a duplicate name, an alias is returned.
        """
        return self._get_model_alias(model) if self.is_duplicate(model) else model.__name__

    def get_field_nullability(self, field: Field) -> bool:
        """Determine if a field requires a value to be provided when instantiating a model.

        In practice, there isn't any reliable way to determine this (even if Django does not provide
        a default, things could be set at the database level). However, we can make some assumptions
        regarding the field instance, see https://forum.djangoproject.com/t/26357 for more details.
        """
        # TODO use the same logic as the mypy plugin
        return (
            field.null
            or field.has_default()
            or getattr(field, "db_default", NOT_PROVIDED) is not NOT_PROVIDED  # No `has_db_default` method :/
            or (isinstance(field, DateField) and (field.auto_now or field.auto_now_add))
        )


def _get_paths_infos(
    url_resolver: URLResolver,
    parent_namespaces: list[str] | None = None,
) -> defaultdict[str, PathInfo]:
    parent_namespaces = parent_namespaces or []
    paths_info: defaultdict[str, PathInfo] = defaultdict(PathInfo)

    for pattern in reversed(url_resolver.url_patterns):  # Parsing in reverse is important!
        if isinstance(pattern, URLPattern) and pattern.name:
            key = ":".join(parent_namespaces)
            if key:
                key += ":"
            key += pattern.name

            reverse_entries = url_resolver.reverse_dict.getlist(pattern.name)

            for possibility, _, defaults, _ in reverse_entries:
                for _, params in possibility:
                    # TODO should `defaults` really be taken into account?
                    # something weird is happening in `_reverse_with_prefix`:
                    # if any(kwargs.get(k, v) != v for k, v in defaults.items()): skip candidate
                    paths_info[key] = paths_info[key].with_new_arguments({k: (k not in defaults) for k in params})
        elif isinstance(pattern, URLResolver):
            new_parent_namespaces = parent_namespaces.copy()
            if pattern.namespace:
                new_parent_namespaces.append(pattern.namespace)

            paths_info = defaultdict(PathInfo, **paths_info, **_get_paths_infos(pattern, new_parent_namespaces))
    return paths_info



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from libcst.codemod import CodemodContext, VisitorBasedCodemodCommand

if TYPE_CHECKING:
    from django_autotyping.app_settings import CodeGenerationSettings

    from ..django_context import DjangoCodemodContext


class BaseVisitorBasedCodemod(VisitorBasedCodemodCommand):
    """The base class for all codemods used for Django user code."""

    def __init__(self, context: CodemodContext) -> None:
        super().__init__(context)
        self.django_context = cast("DjangoCodemodContext", context.scratch["django_context"])
        self.code_generation_settings = cast("CodeGenerationSettings", context.scratch["code_generation_settings"])



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal, Type, TypedDict

import libcst as cst
from django.db import models

from ._compat import TypeAlias

ModelType = Type[models.Model]

if TYPE_CHECKING:
    # See https://github.com/Instagram/LibCST/issues/1075
    FlattenFunctionDef = cst.FlattenSentinel[cst.FunctionDef]
else:
    FlattenFunctionDef = cst.FunctionDef


RulesT: TypeAlias = Literal["DJA001", "DJAS001", "DJAS002", "DJAS010", "DJAS011"]


class AutotypingSettingsDict(TypedDict, total=False):
    """A utility `TypedDict` to be used in user code settings.

    ```python
    AUTOTYPING: AutotypingSettingsDict = {
        "IGNORE": ["DJA001"],
        ...
    }
    ```
    """

    IGNORE: list[RulesT]
    """A list of ignored rules."""

    STUBS_GENERATION: StubsGenerationSettingsDict
    """Stub related settings."""

    CODE_GENERATION: CodeGenerationSettingsDict
    """Code generation related settings."""


class StubsGenerationSettingsDict(TypedDict, total=False):
    """Configuration for dynamic stubs generation."""

    LOCAL_STUBS_DIR: Path | None
    """The directory of the local type stubs. If not set, this setting must be set as a CLI argument."""

    SOURCE_STUBS_DIR: Path | None
    """The directory of the source `django-stubs` to be used. Will default
    to the first entry in site packages.
    """

    ALLOW_PLAIN_MODEL_REFERENCES: bool
    """Whether string references in the form of `{model_name}` should be generated in overloads.

    If set to `True`, both `{model_name}` and `{model_name}.{app_label}` are allowed
    (unless the model name has a duplicate in a different app).

    Affected rules: `DJAS001`.
    """

    ALLOW_NONE_SET_TYPE: bool
    """Whether to allow having the `__set__` type variable set to `None`, even if the field is not nullable.

    While Django allows setting most model instance fields to any value (before saving),
    it is generally a bad practice to do so. However, it might be beneficial to allow `None`
    to be set temporarly.

    This also works for foreign fields, where unlike standard fields, the Django descriptor used
    only allows model instances and `None` to be set.

    Affected rules: `DJAS001`.
    """

    MODEL_FIELDS_OPTIONAL: bool
    """Whether all model fields should be considered optional when creating model instances.

    This affects the following signatures:

    - `Manager.create/acreate`
    - `__init__` methods of models

    A lot can happen behind the scenes when instantiating models. Even if a field doesn't have
    a default value provided, the database could have triggers implemented that would provide one.
    This is why, by default, this configuration attribute defaults to `True`. If set to `False`,
    `django-autotyping` will try its best to determine required fields, namely by checking if:

    - the field can be `null`
    - the field has a default or a database default value set
    - the field is a subclass of `DateField` and has `auto_now` or `auto_now_add` set to `True`.

    Affected rules: `DJAS002`.
    """

    ALLOW_REVERSE_ARGS: bool
    """Whether type checking should be added to the `args` argument of `reverse`.

    By default, this is set to `False` to avoid having too many overloads being generated.
    Moreover, only tuples can be type checked, and most people are using lists for this argument.
    Instead, it is recommended to use the `kwargs` argument.

    Affected rules: `DJAS011`.
    """


class CodeGenerationSettingsDict(TypedDict, total=False):
    """Configuration for adding type annotations to Django user code."""

    PROJECT_DIR: Path | None
    """The directory of your project, where code modifications should be applied.
    If not set, this setting must be set as a CLI argument.
    """

    DIFF: bool
    """Show changes to be applied instead of modifying existing files."""

    TYPE_CHECKING_BLOCK: bool
    """Whether newly added imports should be in an `if TYPE_CHECKING` block (avoids circular imports)."""

    ASSUME_CLASS_GETITEM: bool
    """Whether generic classes in stubs files but not at runtime should be assumed to have a
    `__class_getitem__` method. This can be achieved by using `django-stubs-ext` or manually.

    Affected rules: `DJA001`.
    """



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import libcst as cst
import libcst.matchers as m
from django.db.models import Field
from libcst import helpers
from libcst.codemod import CodemodContext
from libcst.metadata import ScopeProvider

from django_autotyping.typing import FlattenFunctionDef

from .base import InsertAfterImportsVisitor, StubVisitorBasedCodemod
from .constants import OVERLOAD_DECORATOR
from .utils import TypedDictAttribute, build_typed_dict, get_param

if TYPE_CHECKING:
    from ..django_context import DjangoStubbingContext

# Matchers:

MANAGER_QS_CLASS_DEF_MATCHER = m.ClassDef(
    name=m.SaveMatchedNode(m.Name("BaseManager") | m.Name("_QuerySet"), "cls_name")
)
"""Matches the `BaseManager` and `_QuerySet` class definitions."""

MODEL_CLASS_DEF_MATCHER = m.ClassDef(name=m.SaveMatchedNode(m.Name("Model"), "cls_name"))
"""Matches the `Model` class definition."""


CREATE_DEF_MATCHER = m.FunctionDef(name=m.Name("create") | m.Name("acreate"))
"""Matches the `create` and `acreate` method definitions."""


INIT_DEF_MATCHER = m.FunctionDef(name=m.Name("__init__"))
"""Matches the `__init__` method definition."""


class CreateOverloadCodemod(StubVisitorBasedCodemod):
    """A codemod that will add overloads to methods creating an instance of a model.

    **Rule identifier**: `DJAS002`.

    **Related settings**:

    -[`MODEL_FIELDS_OPTIONAL`][django_autotyping.app_settings.StubsGenerationSettings.MODEL_FIELDS_OPTIONAL].

    ```python
    MyModel(...)  # Signature is provided.
    MyModel.objects.create(...)  # Signature is provided.
    ```

    ??? abstract "Implementation"
        This codemod makes use of the [PEP 692][pep-0692]. If your type checker/LSP supports it,
        documentation is provided for each field if [`help_text`][django.db.models.Field.help_text] was set.
    """

    METADATA_DEPENDENCIES = {ScopeProvider}
    STUB_FILES = {"db/models/manager.pyi", "db/models/query.pyi", "db/models/base.pyi"}

    def __init__(self, context: CodemodContext) -> None:
        super().__init__(context)
        self.add_model_imports()
        model_typed_dicts = _build_model_kwargs(self.django_context, self.stubs_settings.MODEL_FIELDS_OPTIONAL)
        InsertAfterImportsVisitor.insert_after_imports(context, model_typed_dicts)

        # Even though these are most likely included, we import them for safety:
        self.add_typing_imports(["TypedDict", "TypeVar", "Unpack", "overload"])

    def mutate_FunctionDef(self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef) -> FlattenFunctionDef:
        cls_name = self.get_metadata(ScopeProvider, original_node).name

        overload = updated_node.with_changes(decorators=[OVERLOAD_DECORATOR])
        overloads: list[cst.FunctionDef] = []

        for model in self.django_context.models:
            model_name = self.django_context.get_model_name(model)

            # sets `self: BaseManager[model_name]/_QuerySet[model_name, _Row]/model_name`
            if cls_name == "_QuerySet":
                annotation = helpers.parse_template_expression(f"{cls_name}[{model_name}, _Row]")
            elif cls_name == "BaseManager":
                annotation = helpers.parse_template_expression(f"{cls_name}[{model_name}]")
            else:
                annotation = helpers.parse_template_expression(model_name)
            self_param = get_param(overload, "self")
            overload_ = overload.with_deep_changes(
                old_node=self_param,
                annotation=cst.Annotation(annotation),
            )

            overload_ = overload_.with_deep_changes(
                old_node=overload_.params.star_kwarg,
                annotation=cst.Annotation(
                    annotation=helpers.parse_template_expression(f"Unpack[{model_name}CreateKwargs]")
                ),
            )

            if cls_name == "Model":
                # Remove `*args` from the definition:
                overload_ = overload_.with_deep_changes(
                    old_node=overload_.params,
                    star_arg=cst.MaybeSentinel.DEFAULT,
                )

            overloads.append(overload_)

        return cst.FlattenSentinel(overloads)

    @m.call_if_inside(MANAGER_QS_CLASS_DEF_MATCHER)
    @m.leave(CREATE_DEF_MATCHER)
    def mutate_create_FunctionDef(
        self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
    ) -> FlattenFunctionDef:
        """Add overloads for `create`/`acreate` if in `BaseManager`/`_QuerSet`."""
        return self.mutate_FunctionDef(original_node, updated_node)

    @m.call_if_inside(MODEL_CLASS_DEF_MATCHER)
    @m.leave(INIT_DEF_MATCHER)
    def mutate_init_FunctionDef(
        self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
    ) -> FlattenFunctionDef:
        """Add overloads for `__init__` if in `Model`."""
        return self.mutate_FunctionDef(original_node, updated_node)


def _build_model_kwargs(django_context: DjangoStubbingContext, all_optional: bool) -> list[cst.ClassDef]:
    class_defs: list[cst.ClassDef] = []

    for model in django_context.models:
        model_name = django_context.get_model_name(model)
        class_defs.append(
            build_typed_dict(
                f"{model_name}CreateKwargs",
                attributes=[
                    TypedDictAttribute(
                        field.name,
                        annotation="Any",
                        docstring=getattr(field, "help_text", None) or None,
                        required=not all_optional and not django_context.get_field_nullability(field),
                    )
                    for field in cast(list[Field], model._meta._get_fields(reverse=False))
                ],
                total=False,
                leading_line=True,
            )
        )

    return class_defs



from __future__ import annotations

from typing import TYPE_CHECKING

import libcst as cst
import libcst.matchers as m
from libcst import helpers
from libcst.codemod import CodemodContext

from django_autotyping.typing import FlattenFunctionDef

from .base import InsertAfterImportsVisitor, StubVisitorBasedCodemod
from .constants import OVERLOAD_DECORATOR
from .utils import TypedDictAttribute, build_typed_dict, get_param

if TYPE_CHECKING:
    from ..django_context import DjangoStubbingContext

# Matchers:

BASE_MANAGER_CLASS_DEF_MATCHER = m.ClassDef(name=m.Name("BaseManager"))
"""Matches the `ManyToManyField` class definition."""

GET_MODEL_DEF_MATCHER = m.FunctionDef(name=m.Name("get"))
"""Matches the `get` method definition."""


class QueryLookupsOverloadCodemod(StubVisitorBasedCodemod):
    """A codemod that will add overloads to the `__init__` methods of related fields.

    Rule identifier: `DJAS003`.
    """

    STUB_FILES = {"db/models/manager.pyi"}

    def __init__(self, context: CodemodContext) -> None:
        super().__init__(context)
        self.add_model_imports()
        model_typed_dicts = _build_model_kwargs(self.django_context.models)
        InsertAfterImportsVisitor.insert_after_imports(context, model_typed_dicts)

        # Even though these are most likely included, we import them for safety:
        self.add_typing_imports(["TypedDict", "TypeVar", "Unpack", "overload"])

    @m.call_if_inside(BASE_MANAGER_CLASS_DEF_MATCHER)
    @m.leave(GET_MODEL_DEF_MATCHER)
    def mutate_classDef(self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef) -> FlattenFunctionDef:
        """Add the necessary overloads to foreign fields that supports
        that supports parametrization of the `__set__` and `__get__` types.
        """

        overload_get = updated_node.with_changes(decorators=[OVERLOAD_DECORATOR])
        overloads: list[cst.FunctionDef] = []

        for model in self.django_context.models:
            model_name = self.django_context.get_model_name(model)

            # sets `self: ManyToManyField[model_name, _Through]`
            self_param = get_param(overload_get, "self")
            overload = overload_get.with_deep_changes(
                old_node=self_param,
                annotation=cst.Annotation(annotation=helpers.parse_template_expression(f"BaseManager[{model_name}]")),
            )

            overload = overload.with_deep_changes(
                old_node=overload.params.star_kwarg,
                annotation=cst.Annotation(annotation=helpers.parse_template_expression(f"Unpack[{model_name}Kwargs]")),
            )

            overloads.append(overload)

        return cst.FlattenSentinel(overloads)


def _build_model_kwargs(django_context: DjangoStubbingContext) -> list[cst.ClassDef]:
    # TODO This needs to build the available lookups
    class_defs: list[cst.ClassDef] = []

    for model in django_context.models:
        model_name = django_context.get_model_name(model)
        class_defs.append(
            build_typed_dict(
                f"{model_name}CreateKwargs",
                attributes=[
                    TypedDictAttribute(
                        field.name,
                        annotation="Any",
                        docstring=getattr(field, "help_text", None) or None,
                    )
                    for field in model._meta._get_fields(reverse=False)
                ],
                total=False,  # TODO find a way to determine which fields are required.
                leading_line=True,
            )
        )

    return class_defs



from __future__ import annotations

import libcst as cst
import libcst.matchers as m
from libcst import helpers
from libcst.codemod import CodemodContext

from django_autotyping.typing import FlattenFunctionDef

from .base import StubVisitorBasedCodemod
from .constants import OVERLOAD_DECORATOR
from .utils import get_param

# Matchers:

CLASS_DEF_MATCHER = m.ClassDef(name=m.Name("Apps"))
"""Matches the `Apps` class definition."""

GET_MODEL_DEF_MATCHER = m.FunctionDef(name=m.Name("get_model"))
"""Matches the `get_model` method definition."""


class GetModelOverloadCodemod(StubVisitorBasedCodemod):
    """A codemod that will add overloads to the [`apps.get_model`][django.apps.apps.get_model] method.

    Rule identifier: `DJAS010`.

    ```python
    reveal_type(apps.get_model("app_name.ModelName"))  # Revealed type is type[ModelName]
    reveal_type(apps.get_model("app_name", "ModelName"))  # Revealed type is type[ModelName]
    ```
    """

    STUB_FILES = {"apps/registry.pyi"}

    def __init__(self, context: CodemodContext) -> None:
        super().__init__(context)
        self.add_model_imports()

        # Even though these are most likely included, we import them for safety:
        self.add_typing_imports(["Literal", "overload"])

    @m.call_if_inside(CLASS_DEF_MATCHER)
    @m.leave(GET_MODEL_DEF_MATCHER)
    def mutate_FunctionDef(self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef) -> FlattenFunctionDef:
        overload = updated_node.with_changes(decorators=[OVERLOAD_DECORATOR])
        overloads: list[cst.FunctionDef] = []

        for model in self.django_context.models:
            for use_shortcut in (True, False):
                model_name = self.django_context.get_model_name(model)
                app_label = model._meta.app_label

                # sets `app_label: Literal[...]`
                app_label_param = get_param(overload, "app_label")
                if use_shortcut:
                    annotation = helpers.parse_template_expression(f'Literal["{app_label}.{model.__name__}"]')
                else:
                    annotation = helpers.parse_template_expression(f'Literal["{app_label}"]')
                overload_ = overload.with_deep_changes(
                    old_node=app_label_param,
                    annotation=cst.Annotation(annotation),
                )

                # sets `model_name: Literal[...]`
                model_name_param = get_param(overload_, "model_name")
                if use_shortcut:
                    annotation = helpers.parse_template_expression("Literal[None]")
                else:
                    annotation = helpers.parse_template_expression(f'Literal["{model.__name__}"]')

                overload_ = overload_.with_deep_changes(
                    old_node=model_name_param,
                    annotation=cst.Annotation(annotation),
                    default=None if not use_shortcut else model_name_param.default,
                    equal=cst.MaybeSentinel.DEFAULT if not use_shortcut else model_name_param.equal,
                )

                # sets return value
                overload_ = overload_.with_changes(
                    # This time use the imported model name!
                    returns=cst.Annotation(helpers.parse_template_expression(f"type[{model_name}]"))
                )

                overloads.append(overload_)

        return cst.FlattenSentinel(overloads)



from __future__ import annotations

import inspect
from pathlib import Path
from typing import TYPE_CHECKING

from django_autotyping._compat import is_relative_to

from .models import ModelInfo

if TYPE_CHECKING:
    from django.apps.registry import Apps
    from django.conf import LazySettings


class DjangoCodemodContext:
    def __init__(self, apps: Apps, settings: LazySettings, project_dir: Path) -> None:
        self.apps = apps
        self.settings = settings
        self.project_dir = project_dir

    @property
    def model_infos(self) -> list[ModelInfo]:
        """A list of `ModelInfo` objects.

        Only the models defined in files relative to `self.project_dir` will be taken into account.
        """
        return [
            ModelInfo.from_model(model)
            for model in self.apps.get_models()
            if is_relative_to(Path(inspect.getabsfile(model)), self.project_dir.resolve())
        ]



from __future__ import annotations

import shutil
import site
from pathlib import Path

import libcst as cst
from libcst.codemod import CodemodContext

from django_autotyping.app_settings import StubsGenerationSettings

from .codemods import StubVisitorBasedCodemod
from .django_context import DjangoStubbingContext


def run_codemods(
    codemods: list[type[StubVisitorBasedCodemod]],
    django_context: DjangoStubbingContext,
    stubs_settings: StubsGenerationSettings,
) -> None:
    django_stubs_dir = stubs_settings.SOURCE_STUBS_DIR or _get_django_stubs_dir()

    for codemod in codemods:
        for stub_file in codemod.STUB_FILES:
            context = CodemodContext(
                filename=stub_file, scratch={"django_context": django_context, "stubs_settings": stubs_settings}
            )
            transformer = codemod(context)
            source_file = django_stubs_dir / stub_file
            target_file = stubs_settings.LOCAL_STUBS_DIR / "django-stubs" / stub_file

            input_code = source_file.read_text(encoding="utf-8")
            input_module = cst.parse_module(input_code)
            output_module = transformer.transform_module(input_module)

            target_file.write_text(output_module.code, encoding="utf-8")


def _get_django_stubs_dir() -> Path:
    # TODO should we use importlib.metadata.files instead?
    for dir in site.getsitepackages():
        if (path := Path(dir, "django-stubs")).is_dir():
            return path
    raise RuntimeError("Couldn't find 'django-stubs' in any of the site packages.")


def create_local_django_stubs(stubs_dir: Path, source_django_stubs: Path | None = None) -> None:
    """Copy the `django-stubs` package into the specified local stubs directory.

    If `source_django_stubs` is not provided, the first entry in site packages will be used.
    """
    stubs_dir.mkdir(exist_ok=True)
    source_django_stubs = source_django_stubs or _get_django_stubs_dir()
    if not (stubs_dir / "django-stubs").is_dir():
        shutil.copytree(source_django_stubs, stubs_dir / "django-stubs")

    # for stub_file in django_stubs_dir.glob("**/*.pyi"):
    #     # Make file relative to site packages, results in `Path("django-stubs/path/to/file.pyi")`
    #     relative_stub_file = stub_file.relative_to(django_stubs_dir.parent)
    #     symlinked_path = stubs_dir / relative_stub_file

    #     stub_file.mkdir()



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Sequence, TypeVar, cast

import libcst as cst
import libcst.matchers as m
from libcst.codemod import CodemodContext, ContextAwareTransformer, VisitorBasedCodemodCommand
from libcst.codemod.visitors import AddImportsVisitor

if TYPE_CHECKING:
    from django_autotyping.app_settings import StubsGenerationSettings

    from ..django_context import DjangoStubbingContext


TYPING_EXTENSIONS_NAMES = ["Unpack", "Required", "NotRequired"]

ModuleT = TypeVar("ModuleT", bound=cst.Module)


IMPORT_MATCHER = m.SimpleStatementLine(body=[m.Import() | m.ImportFrom() | m.ImportAlias() | m.ImportStar()])
"""Matches the definition of an import statement."""


class InsertAfterImportsVisitor(ContextAwareTransformer):
    """Insert a list of statements after imports."""

    CONTEXT_KEY = "InsertAfterImportsVisitor"

    @classmethod
    def insert_after_imports(
        cls,
        context: CodemodContext,
        statements: Sequence[cst.SimpleStatementLine | cst.BaseCompoundStatement],
    ) -> None:
        """Insert a list of statements following the module imports.

        If no imports are to be found, statements will be added at the beginning of the module."""
        ctx_statements = context.scratch.get(cls.CONTEXT_KEY, [])
        ctx_statements.extend(statements)
        context.scratch[cls.CONTEXT_KEY] = ctx_statements

    def leave_Module(self, original_node: cst.Module, updated_node: cst.Module) -> cst.Module:
        statements = self.context.scratch.get(self.CONTEXT_KEY, [])
        if not statements:
            return updated_node

        body = list(updated_node.body)

        last_import = next((node for node in reversed(body) if m.matches(node, IMPORT_MATCHER)), None)
        index = body.index(last_import) + 1 if last_import is not None else 0
        body[index:index] = statements

        return updated_node.with_changes(
            body=body,
        )


class StubVisitorBasedCodemod(VisitorBasedCodemodCommand):
    """The base class for all codemods used for custom stub files."""

    STUB_FILES: ClassVar[set[str]]

    def __init__(self, context: CodemodContext) -> None:
        super().__init__(context)
        self.django_context = cast("DjangoStubbingContext", context.scratch["django_context"])
        self.stubs_settings = cast("StubsGenerationSettings", context.scratch["stubs_settings"])

    def add_model_imports(self) -> None:
        """Add the defined models in the Django context as imports to the current file."""

        # TODO LibCST should support adding imports from `ImportItem` objects
        imports = AddImportsVisitor._get_imports_from_context(self.context)
        imports.extend(self.django_context.model_imports)
        self.context.scratch[AddImportsVisitor.CONTEXT_KEY] = imports

    def add_typing_imports(self, names: list[str]) -> None:
        """Add imports to the `typing` module (either from `typing` or `typing_extensions`)."""
        for name in names:
            AddImportsVisitor.add_needed_import(
                self.context,
                module="typing_extensions" if name in TYPING_EXTENSIONS_NAMES else "typing",
                obj=name,
            )

    def transform_module(self, tree: cst.Module) -> cst.Module:
        # LibCST automatically runs `AddImportsVisitor` and `RemoveImportsVisitor`,
        # but this is hardcoded. So we manually add our visitor.
        tree = super().transform_module(tree)
        transform = InsertAfterImportsVisitor

        return self._instantiate_and_run(transform, tree)



from __future__ import annotations

__all__ = ("ForwardRelationTypingCodemod", "RulesT", "rules", "gather_codemods")

from typing import Container, Literal

from django_autotyping._compat import TypeAlias

from .base import BaseVisitorBasedCodemod
from .forward_relation_typing_codemod import ForwardRelationTypingCodemod

RulesT: TypeAlias = Literal["DJA001"]

rules: list[tuple[RulesT, type[BaseVisitorBasedCodemod]]] = [
    ("DJA001", ForwardRelationTypingCodemod),
]


def gather_codemods(
    ignore: Container[RulesT] = [], include: Container[RulesT] = []
) -> list[type[BaseVisitorBasedCodemod]]:
    if include:
        return [rule[1] for rule in rules if rule[0] in include]
    return [rule[1] for rule in rules if rule[0] not in ignore]



from __future__ import annotations

import inspect
from dataclasses import dataclass, field
from types import ModuleType
from typing import cast

from django.apps import AppConfig, apps
from django.db.models.fields.related import RelatedField

from ..typing import ModelType


@dataclass
class ForwardRelation:
    class_name: str
    nullable: bool
    has_class_getitem: bool
    model: ModelType

    @property
    def model_module(self) -> ModuleType:
        """The module object of the model linked to this relation."""
        return cast(ModuleType, inspect.getmodule(self.model))

    @property
    def app_config(self) -> AppConfig:
        """The `AppConfig` object where the model linked to this relation belongs."""
        return apps.get_app_config(self.model._meta.app_label)

    @property
    def app_models_module(self) -> ModuleType | None:
        """The module object where models of the app of the model linked to this relation are stored."""
        return self.app_config.models_module  # type: ignore[return-value]

    @classmethod
    def from_field(cls, field: RelatedField) -> ForwardRelation:
        return cls(
            class_name=field.__class__.__name__,
            nullable=field.null,
            has_class_getitem=hasattr(type(field), "__class_getitem__"),
            model=field.related_model,
        )


@dataclass
class ModelInfo:
    model: ModelType
    module: ModuleType
    forward_relations: dict[str, ForwardRelation] = field(default_factory=dict)

    @property
    def class_name(self) -> str:
        return self.model.__name__

    @property
    def filename(self) -> str:
        return cast(str, inspect.getsourcefile(self.model))

    @property
    def app_label(self) -> str:
        """The app label where the model belongs."""
        return self.model._meta.app_label

    @property
    def app_config(self) -> AppConfig:
        """The `AppConfig` object where the model belongs."""
        return self.model._meta.app_config

    @classmethod
    def from_model(cls, model: ModelType) -> ModelInfo:
        forward_relations = {
            field.name: ForwardRelation.from_field(field)
            for field in model._meta.get_fields()
            if isinstance(field, RelatedField)  # TODO isinstance check on `Field`?
            # if field.many_to_one  # TODO may be unnecessary?
        }

        return cls(
            model=model,
            module=inspect.getmodule(model),
            forward_relations=forward_relations,
        )
